import { useLayoutEffect, useEffect, useState, useRef, useMemo, createElement, useCallback } from 'react';
import { createContext, useContext, useContextSelector } from 'use-context-selector';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const isClient = typeof window !== 'undefined' && !/ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
const useIsoLayoutEffect = isClient ? useLayoutEffect : useEffect;
const appendMap = (dst, src) => {
  src.forEach((v, k) => {
    dst.set(k, v);
  });
  return dst;
};
const concatMap = (src1, src2) => {
  const dst = new Map();
  src1.forEach((v, k) => {
    dst.set(k, v);
  });
  src2.forEach((v, k) => {
    dst.set(k, v);
  });
  return dst;
};

const warningObject = new Proxy({}, {
  get() {
    throw new Error('Please use <Provider>');
  },

  apply() {
    throw new Error('Please use <Provider>');
  }

});

const addDependent = (dependentsMap, atom, dependent) => {
  let dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

const deleteDependent = (dependentsMap, atom, dependent) => {
  const dependents = dependentsMap.get(atom);

  if (dependents && dependents.has(dependent)) {
    dependents.delete(dependent);
    return dependents.size === 0; // empty
  }

  return false; // not found
};

const listDependents = (dependentsMap, atom) => {
  const dependents = dependentsMap.get(atom);
  return dependents || new Set();
};

const initialState = new Map();

const getAtomState = (state, atom) => {
  const atomState = state.get(atom);

  if (!atomState) {
    throw new Error('atom is not initialized');
  }

  return atomState;
};

const getAtomStateValue = (state, atom) => {
  const atomState = state.get(atom);
  return atomState ? atomState.value : atom.initialValue;
};

const initAtom = (id, initializingAtom, stateRef, setState, dependentsMap) => {
  const createAtomState = (prevState, atom, dependent) => {
    addDependent(dependentsMap, atom, dependent);
    const partialState = new Map();
    let atomState = prevState.get(atom);

    if (atomState) {
      return partialState; // already initialized
    }

    let isSync = true;
    const nextValue = atom.read(a => {
      if (a !== atom) {
        if (isSync) {
          const nextPartialState = createAtomState(prevState, a, atom);
          appendMap(partialState, nextPartialState);
        } else {
          const nextPartialState = createAtomState(stateRef.current, a, atom);
          setState(prev => appendMap(new Map(prev), nextPartialState));
        }
      }

      return getAtomStateValue(prevState, a);
    });

    if (nextValue instanceof Promise) {
      const promise = nextValue.then(value => {
        setState(prev => new Map(prev).set(atom, {
          promise: null,
          value
        }));
      });
      atomState = {
        promise,
        value: atom.initialValue
      };
    } else {
      atomState = {
        promise: null,
        value: nextValue
      };
    }

    partialState.set(atom, atomState);
    isSync = false;
    return partialState;
  };

  const updateState = createAtomState(stateRef.current, initializingAtom, id);
  setState(prevState => appendMap(new Map(prevState), updateState));
};

const disposeAtom = (id, setState, dependentsMap) => {
  const deleteAtomState = (prevState, dependent) => {
    let nextState = new Map(prevState);
    const deleted = [];
    nextState.forEach((_atomState, atom) => {
      const isEmpty = deleteDependent(dependentsMap, atom, dependent);

      if (isEmpty) {
        nextState.delete(atom);
        deleted.push(atom); // TODO delete in dependentsMap too (even though it is WeakMap)
      }
    });
    nextState = deleted.reduce((p, c) => deleteAtomState(p, c), nextState);
    return nextState;
  };

  setState(prevState => deleteAtomState(prevState, id));
};

const updateAtomValue = (updatingAtom, update, stateRef, setState, dependentsMap) => {
  const updateDependentsState = (prevState, atom) => {
    const partialState = new Map();
    listDependents(dependentsMap, atom).forEach(dependent => {
      if (typeof dependent === 'symbol') return;
      const v = dependent.read(a => {
        if (a !== dependent) {
          addDependent(dependentsMap, a, dependent);
        }

        return getAtomStateValue(prevState, a);
      });

      if (v instanceof Promise) {
        const promise = v.then(vv => {
          const nextAtomState = {
            promise: null,
            value: vv
          };
          const nextPartialState = updateDependentsState(new Map(stateRef.current).set(dependent, nextAtomState), dependent);
          setState(prev => appendMap(new Map(prev).set(dependent, {
            promise: null,
            value: vv
          }), nextPartialState));
        });
        partialState.set(dependent, _extends({}, getAtomState(prevState, dependent), {
          promise
        }));
      } else {
        partialState.set(dependent, {
          promise: null,
          value: v
        });
        appendMap(partialState, updateDependentsState(prevState, dependent));
      }
    });
    return partialState;
  };

  const updateAtomState = (prevState, atom, value) => {
    const partialState = new Map();
    let isSync = true;
    const promise = atom.write(a => getAtomStateValue(prevState, a), (a, v) => {
      if (a === atom) {
        const nextAtomState = {
          promise: null,
          value: v
        };

        if (isSync) {
          partialState.set(a, nextAtomState);
          appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), a));
        } else {
          const nextPartialState = updateDependentsState(new Map(stateRef.current).set(a, nextAtomState), a);
          setState(prev => appendMap(new Map(prev).set(a, nextAtomState), nextPartialState));
        }
      } else {
        if (isSync) {
          const nextPartialState = updateAtomState(concatMap(prevState, partialState), a, v);
          appendMap(partialState, nextPartialState);
        } else {
          const nextPartialState = updateAtomState(stateRef.current, a, v);
          setState(prev => appendMap(new Map(prev), nextPartialState));
        }
      }
    }, value);

    if (promise instanceof Promise) {
      const nextAtomState = _extends({}, getAtomState(prevState, atom), {
        promise: promise.then(() => {
          setState(prev => new Map(prev).set(atom, _extends({}, getAtomState(prev, atom), {
            promise: null
          })));
        })
      });

      partialState.set(atom, nextAtomState);
    }

    isSync = false;
    return partialState;
  };

  const updatingAtomState = getAtomState(stateRef.current, updatingAtom);

  if (updatingAtomState.promise) {
    // schedule update after promise is resolved
    const promise = updatingAtomState.promise.then(() => {
      const updateState = updateAtomState(stateRef.current, updatingAtom, typeof update === 'function' ? update(getAtomStateValue(stateRef.current, updatingAtom)) : update);
      setState(prevState => appendMap(new Map(prevState), updateState));
    });
    setState(prevState => new Map(prevState).set(updatingAtom, _extends({}, updatingAtomState, {
      promise
    })));
  } else {
    const updateState = updateAtomState(stateRef.current, updatingAtom, typeof update === 'function' ? update(updatingAtomState.value) : update);
    setState(prevState => appendMap(new Map(prevState), updateState));
  }
};

const ActionsContext = createContext(warningObject);
const StateContext = createContext(warningObject);
const Provider = ({
  children
}) => {
  const [state, setState] = useState(initialState);
  const stateRef = useRef(state);
  useIsoLayoutEffect(() => {
    stateRef.current = state;
  });
  const dependentsMapRef = useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new WeakMap();
  }

  const actions = useMemo(() => ({
    init: (id, atom) => initAtom(id, atom, stateRef, setState, dependentsMapRef.current),
    dispose: id => disposeAtom(id, setState, dependentsMapRef.current),
    update: (atom, update) => updateAtomValue(atom, update, stateRef, setState, dependentsMapRef.current)
  }), []);
  return createElement(ActionsContext.Provider, {
    value: actions
  }, createElement(StateContext.Provider, {
    value: state
  }, children));
};

function atom(read, write) {
  const instance = {
    initialValue: null
  };

  if (typeof read === 'function') {
    instance.read = read;
    const value = read(a => a.initialValue);

    if (value instanceof Promise) {
      value.then(v => {
        instance.initialValue = v;
      });
    } else {
      instance.initialValue = value;
    }
  } else {
    instance.initialValue = read;

    instance.read = get => get(instance);

    instance.write = (_get, set, writeValue) => {
      set(instance, writeValue);
    };
  }

  if (write) {
    instance.write = write;
  }

  return instance;
}

const isWritable = atom => !!atom.write;

function useAtom(atom) {
  const actions = useContext(ActionsContext);
  const promiseOrValue = useContextSelector(StateContext, useCallback(state => {
    const atomState = state.get(atom);
    if (!atomState) return atom.initialValue;
    if (atomState.promise) return atomState.promise;
    return atomState.value;
  }, [atom]));
  const setAtom = useCallback(update => {
    if (isWritable(atom)) {
      actions.update(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  useIsoLayoutEffect(() => {
    const id = Symbol();
    actions.init(id, atom);
    return () => {
      actions.dispose(id);
    };
  }, [actions, atom]);

  if (promiseOrValue instanceof Promise) {
    throw promiseOrValue;
  }

  return [promiseOrValue, setAtom];
}

export { Provider, atom, useAtom };
